# Generated by Django 3.1.14 on 2022-04-12 13:13

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('fees', '0003_auto_20220412_1512'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='packagequota',
            unique_together={('package', 'quota')},
        ),
    ]



import django_rq
from django.apps import AppConfig
from django.conf import settings
from django.utils.translation import ugettext_lazy as _

from fees import settings as fees_settings
from fees.cron import send_subscription_reminders


class Config(AppConfig):
    name = 'fees'
    verbose_name = _('Fees')

    def schedule_jobs(self):
        scheduler = django_rq.get_scheduler('cron')

        if fees_settings.SCHEDULE_SUBSCRIPTION_REMINDERS:
            # Cron task to send subscription reminders
            scheduler.cron(
                "0 9 * * *",  # Run every day at 9:00 [UTC]
                # "* * * * *",  # Run every minute
                func=send_subscription_reminders,
                timeout=settings.RQ_QUEUES['cron']['DEFAULT_TIMEOUT']
            )



import datetime
from django.db.models import F
from django.utils.translation import ugettext_lazy as _


class PurchaserMixin(object):
    created_attribute = 'created'

    def is_package_quota_available(self, quota):
        package = self.package
        return quota in package.get_quotas() if package is not None else False

    @property
    def package(self):
        from fees.models import Package
        return Package.get_current_package(self)

    @property
    def plan(self):
        return self.plan_history\
            .active()\
            .order_by(
                '-modified',
                # F('expiration').desc(nulls_last=True)
            ).first()

    @plan.setter
    def plan(self, plan):
        if plan.purchaser and plan.purchaser != self:
            raise ValueError(_('Purchaser of the plan is already set: %s') % plan.purchaser)
        plan.purchaser = self
        plan.save()

    @property
    def plan_history(self):
        return self.plan_set.all()

    @property
    def is_in_free_trial(self):
        if not self.package or not self.plan:
            return False

        trial_duration = self.package.trial_duration

        if not trial_duration or trial_duration <= 0:
            return False

        created_date = getattr(self, self.created_attribute, None)
        plan_expiration_date = self.plan.expiration

        if created_date is not None and plan_expiration_date is not None:
            created_date = created_date.date() if isinstance(created_date, datetime.datetime) else created_date
            delta = plan_expiration_date - created_date

            if delta.days <= trial_duration:
                return True

        return False



from django.urls import path
from django.utils.translation import pgettext_lazy
from fees.views import PackageListView, PlanDetailView

app_name = 'fees'

urlpatterns = [
    path(pgettext_lazy("url", 'pricing/'), PackageListView.as_view(), name='pricing'),
    path(pgettext_lazy("url", 'plan/'), PlanDetailView.as_view(), name='plan'),
]



# Generated by Django 3.1.14 on 2022-04-04 19:28

from django.db import migrations, models
import modeltrans.fields


class Migration(migrations.Migration):

    dependencies = [
        ('fees', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='package',
            name='description',
            field=models.TextField(blank=True, verbose_name='description'),
        ),
        migrations.AddField(
            model_name='package',
            name='i18n',
            field=modeltrans.fields.TranslationField(fields=('description',), required_languages=(), virtual_fields=True),
        ),
        migrations.AlterField(
            model_name='package',
            name='is_visible',
            field=models.BooleanField(db_index=True, default=True, help_text='Is visible in pricing page', verbose_name='visible'),
        ),
        migrations.AlterField(
            model_name='package',
            name='title',
            field=models.CharField(max_length=50, unique=True, verbose_name='title'),
        ),
    ]



from django_rq import job


@job('default')
def send_subscription_reminders(threshold):
    from fees.models import Plan

    expiring_plans = Plan.objects.expires_in(days=threshold)
    total_plans = expiring_plans.count()
    print(f'Found {total_plans} expiring plans in {threshold} days')

    for plan in expiring_plans:
        print(plan, plan.pricing, plan.expiration)
        plan.send_reminder()



from django.views.generic import ListView, DetailView

from fees.models import Package, Plan, Pricing


class PackageListView(ListView):
    model = Package

    def get_queryset(self):
        return super().get_queryset().visible()

    def get_context_data(self, *args, **kwargs):
        context_data = super().get_context_data(*args, **kwargs)
        context_data.update({'distinct_period_durations': Pricing.distinct_period_durations()})
        return context_data


class PlanDetailView(DetailView):
    model = Plan

    # TODO: get purchaser
    def get_object(self, queryset=None):
        try:
            return self.request.user.subscription
        except AttributeError:
            return None



from django.apps import apps as django_apps
from django.core.exceptions import ImproperlyConfigured

from fees import settings as fees_settings


def get_purchaser_model():
    """
    Return the model that represents purchaser object.
    """
    try:
        return django_apps.get_model(fees_settings.PURCHASER_MODEL, require_ready=False)
    except ValueError:
        raise ImproperlyConfigured("FEES_PURCHASER_MODEL must be of the form 'app_label.model_name'")
    except LookupError:
        raise ImproperlyConfigured(
            "FEES_PURCHASER_MODEL refers to model '%s' that has not been installed" % fees_settings.PURCHASER_MODEL
        )



# Generated by Django 3.1.14 on 2023-02-22 23:08

from django.db import migrations, models
import modeltrans.fields


class Migration(migrations.Migration):

    dependencies = [
        ('fees', '0009_auto_20220706_1719'),
    ]

    operations = [
        migrations.AlterField(
            model_name='package',
            name='i18n',
            field=modeltrans.fields.TranslationField(fields=('title', 'description'), required_languages=(), virtual_fields=True),
        ),
        migrations.AlterField(
            model_name='plan',
            name='expiration',
            field=models.DateField(blank=True, db_index=True, default=None, help_text='keep blank to calculate automatically by pricing period', null=True, verbose_name='expires'),
        ),
        migrations.AlterUniqueTogether(
            name='pricing',
            unique_together={('package', 'period', 'duration')},
        ),
    ]



from django.conf import settings

PURCHASER_MODEL = getattr(settings, 'FEES_PURCHASER_MODEL', settings.AUTH_USER_MODEL)
CURRENCY = getattr(settings, 'FEES_CURRENCY', 'EUR')
SCHEDULE_SUBSCRIPTION_REMINDERS = getattr(settings, 'FEES_SCHEDULE_SUBSCRIPTION_REMINDERS', False)
MIGRATION_DEPENDENCIES = getattr(settings, 'FEES_MIGRATION_DEPENDENCIES', [])



from django.core.exceptions import ObjectDoesNotExist

from fees.models import Package


def purchaser_plan(request):
    # let's presume the purchaser is set in the request, otherwise use request.user as purchaser instance
    purchaser = getattr(request, 'purchaser', request.user)

    try:
        plan = purchaser.plan
        package = purchaser.plan.package
    except (ObjectDoesNotExist, AttributeError):
        plan = None
        package = Package.get_fallback_package()

    return {
        'plan': plan,
        'package': package,
        'expiration': plan.expiration if plan else None
    }



from django import template

from fees.models import Quota, Package

register = template.Library()


@register.simple_tag
def package_quotas_info(package):
    quotas = Quota.objects.values('codename', 'name_i18n', 'description_i18n', 'is_boolean')
    all_quotas = dict((
        q['codename'],
        {
            'name': q['name_i18n'],
            'description': q['description_i18n'],
            'is_boolean': q['is_boolean'],
        }
    ) for q in quotas)

    quotas_info = all_quotas.copy()
    package_quotas = package.packagequota_set.all().values('quota__codename', 'value')

    for package_quota in package_quotas:
        codename = package_quota['quota__codename']
        quota = quotas_info[codename]

        quota.update({'is_available': Package.is_quota_available(package_quota)})

        if not quota['is_boolean']:
            quota.update({
                'value': package_quota['value'],
            })

    return quotas_info



# Generated by Django 3.1.14 on 2022-04-12 13:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('fees', '0002_auto_20220404_2128'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='pricing',
            options={'ordering': ['price'], 'verbose_name': 'pricing', 'verbose_name_plural': 'pricings'},
        ),
        migrations.AlterModelOptions(
            name='quota',
            options={'ordering': ('order',), 'verbose_name': 'Quota', 'verbose_name_plural': 'Quotas'},
        ),
        migrations.AddField(
            model_name='quota',
            name='order',
            field=models.PositiveSmallIntegerField(default=1, help_text='to set order in pricing', verbose_name='ordering'),
        ),
        migrations.AlterUniqueTogether(
            name='pricing',
            unique_together={('package', 'period')},
        ),
    ]



from datetime import timedelta

from django.db import models
from django.db.models import Q
from django.utils.timezone import now


class PackageQuerySet(models.QuerySet):
    def visible(self):
        return self.filter(is_visible=True)


class PlanQuerySet(models.QuerySet):
    def active(self):
        return self.filter(
            Q(activation__lte=now()),
            Q(expiration=None) | Q(expiration__gte=now()),
        )

    def expires_in(self, days=7):
        threshold = now() + timedelta(days=days)
        return self.filter(expiration=threshold.date())



# Generated by Django 3.1.14 on 2022-07-06 15:07

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('fees', '0007_auto_20220706_1652'),
    ]

    operations = [
        migrations.AddField(
            model_name='plan',
            name='activation',
            field=models.DateField(auto_now_add=True, default=django.utils.timezone.now, verbose_name='activation'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='package',
            name='is_default',
            field=models.BooleanField(db_index=True, default=False, help_text='Default package for new purchaser (useful for trial packages). Only 1 default package at a time can be set.', verbose_name='default'),
        ),
        migrations.AlterField(
            model_name='package',
            name='is_fallback',
            field=models.BooleanField(db_index=True, default=False, help_text='Fallback package for purchaser when its subscription expires or trial ends (useful for freemium packages). Only 1 fallback package at a time can be set.', verbose_name='fallback'),
        ),
    ]



from copy import deepcopy

from django.contrib import admin, messages
from django.contrib.admin.options import csrf_protect_m
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _
from django.utils.html import format_html

from fees.helpers import get_purchaser_model
from fees.models import Plan, Package, PackageQuota, Quota, Pricing


class PurchaserLinkMixin(object):
    def purchaser_link(self, obj):
        purchaser_model = get_purchaser_model()
        app_label = purchaser_model._meta.app_label
        model_name = purchaser_model._meta.model_name
        change_url = reverse('admin:%s_%s_change' % (app_label, model_name), args=(obj.purchaser.id,))
        return format_html('<a href="{}">{}</a>', change_url, obj.user.username)

    purchaser_link.short_description = _('Purchaser')
    purchaser_link.allow_tags = True


class QuotaInline(admin.TabularInline):
    model = PackageQuota


class PricingInline(admin.TabularInline):
    model = Pricing


class QuotaAdmin(admin.ModelAdmin):
    list_display = [
        'codename', 'name_i18n', 'description_i18n', 'unit',
        'is_boolean', 'order'
        #'move_up_down_links',
    ]
    list_display_links = ['codename']
    list_editable = ['order']


def copy_package(modeladmin, request, queryset):
    """
    Admin command for duplicating plans preserving quotas and pricings.
    """
    for package in queryset:
        package_copy = deepcopy(package)
        package_copy.id = None
        package_copy.is_available = False
        package_copy.is_default = False
        package_copy.is_fallback = False
        package_copy.trial_duration = 0
        package_copy.created = None
        package_copy.save(force_insert=True)

        for pricing in package.pricing_set.all():
            pricing.id = None
            pricing.plan = package_copy
            pricing.save(force_insert=True)

        for quota in package.packagequota_set.all():
            quota.id = None
            quota.package = package_copy
            quota.save(force_insert=True)


copy_package.short_description = _('Duplicate package')


@admin.register(Package)
class PackageAdmin(admin.ModelAdmin):
    search_fields = ('title_i18n',
                     # 'customized__username', 'customized__email',
                     )
    list_display = [
        'title_i18n',
        # 'description',
        # 'customized',
        'trial_duration',
        'is_default', 'is_fallback', 'is_available', 'is_visible',
        'created',
        'order'
        # 'move_up_down_links'
    ]
    list_editable = ('is_default', 'is_fallback', 'is_available', 'is_visible', 'order')
    list_filter = ('is_default', 'is_fallback', 'is_available', 'is_visible')
    inlines = (
        PricingInline,
        QuotaInline,
    )
    list_select_related = True
    # raw_id_fields = ('customized',)
    actions = [copy_package, ]

    # def queryset(self, request):
    #     return super(PlanAdmin, self).queryset(request).select_related(
    #         'customized'
    #     )

    @csrf_protect_m
    def changelist_view(self, request, extra_context=None):
        from django.db.utils import IntegrityError
        try:
            return super().changelist_view(request, extra_context)
        except IntegrityError as e:
            messages.error(request, str(e))
            return redirect(request.get_full_path())

    def save_model(self, request, obj, form, change):
        if obj.is_default:
            self.model.objects.filter(is_default=True).exclude(id=obj.id).update(is_default=False)
        if obj.is_fallback:
            self.model.objects.filter(is_fallback=True).exclude(id=obj.id).update(is_fallback=False)
        return super().save_model(request, obj, form, change)


# class RecurringPlanInline(admin.StackedInline):
#     model = RecurringUserPlan
#     extra = 0


class PricingAdmin(admin.ModelAdmin):
    list_select_related = ['package']
    list_display = [
        'id',
        'package',
        'duration', 'period',
        'price',
        'timedelta',
    ]


@admin.register(Plan)
class PlanAdmin(PurchaserLinkMixin, admin.ModelAdmin):
    actions = ['send_reminder']
    list_filter = (
        #'is_active',
        'activation', 'expiration', 'package__title', 'package__is_available', 'package__is_visible', 'modified', 'pricing')
    search_fields = ('purchaser__email', 'package__title',)
    list_display = ('id', 'purchaser', 'package', 'pricing', 'activation', 'expiration',  # 'is_active',
                    # 'recurring__automatic_renewal', 'recurring__pricing'
                    'modified')
    list_select_related = True
    readonly_fields = ['purchaser_link', ]
    # inlines = (RecurringPlanInline,)
    fields = ('purchaser', 'package', 'pricing', 'expiration', )  #'is_active')
    autocomplete_fields = ['purchaser', 'package', ]

    def recurring__automatic_renewal(self, obj):
        return obj.recurring.has_automatic_renewal
    recurring__automatic_renewal.admin_order_field = 'recurring__has_automatic_renewal'
    recurring__automatic_renewal.boolean = True

    def recurring__pricing(self, obj):
        return obj.recurring.pricing
    recurring__automatic_renewal.admin_order_field = 'recurring__pricing'

    def send_reminder(self, request, queryset):
        for obj in queryset:
            if obj.is_expired():
                messages.error(request, _('Plan %s already expired') % obj)
            else:
                messages.success(request, _('Reminder sent (%s)') % obj)
                obj.send_reminder()
    send_reminder.short_description = _('Send reminder')


admin.site.register(Quota, QuotaAdmin)
# admin.site.register(Pricing, PricingAdmin)



# Generated by Django 3.1.14 on 2022-07-06 15:19

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('fees', '0008_auto_20220706_1707'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='plan',
            options={'get_latest_by': 'expiration', 'ordering': ['-activation'], 'verbose_name': 'Plan', 'verbose_name_plural': 'Plans'},
        ),
        migrations.AlterField(
            model_name='plan',
            name='purchaser',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='provisioning.workspace', verbose_name='purchaser'),
        ),
    ]



import logging
from datetime import date, timedelta

from dateutil.relativedelta import relativedelta
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.conf import settings
from django.core.validators import MinValueValidator, EMPTY_VALUES
from django.urls import reverse
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _, override as override_language
from modeltrans.fields import TranslationField

from pragmatic.managers import EmailManager

from fees.querysets import PackageQuerySet, PlanQuerySet
from fees import settings as fees_settings
from .helpers import get_purchaser_model

logger = logging.getLogger('fees')


class Quota(models.Model):
    """
    Single countable or boolean property of system (limitation).
    """
    codename = models.CharField(_('codename'), max_length=50, unique=True, db_index=True)
    name = models.CharField(_('name'), max_length=100)
    unit = models.CharField(_('unit'), max_length=100, blank=True)
    description = models.TextField(_('description'), blank=True)
    is_boolean = models.BooleanField(_('is boolean'), default=False)
    order = models.PositiveSmallIntegerField(verbose_name=_('ordering'), help_text=_('to set order in pricing'), default=1)
    i18n = TranslationField(fields=('name', 'description'))

    class Meta:
        ordering = ('order',)
        verbose_name = _("Quota")
        verbose_name_plural = _("Quotas")

    def __str__(self):
        return "%s" % (self.name_i18n, )


class PackageQuotaManager(models.Manager):
    def get_query_set(self):
        return super(PackageQuotaManager, self).get_query_set().select_related('package', 'quota')


class PackageQuota(models.Model):
    package = models.ForeignKey('Package', on_delete=models.CASCADE)
    quota = models.ForeignKey('Quota', on_delete=models.CASCADE)
    value = models.IntegerField(default=1, null=True, blank=True)
    objects = PackageQuotaManager()

    class Meta:
        verbose_name = _("Package quota")
        verbose_name_plural = _("Packages quotas")
        unique_together = (('package', 'quota'),)


class Package(models.Model):
    title = models.CharField(_('title'), unique=True, max_length=50)
    description = models.TextField(_('description'), blank=True)
    order = models.PositiveSmallIntegerField(verbose_name=_('ordering'), help_text=_('to set order in pricing'), unique=True, default=1)
    trial_duration = models.PositiveSmallIntegerField(verbose_name=_('trial duration'), help_text=_('in days'), default=0)
    is_default = models.BooleanField(
        _('default'),
        help_text=_('Default package for new purchaser (useful for trial packages). Only 1 default package at a time can be set.'),
        default=False,
        db_index=True,
    )
    is_fallback = models.BooleanField(
        _('fallback'),
        help_text=_('Fallback package for purchaser when its subscription expires or trial ends (useful for freemium packages). Only 1 fallback package at a time can be set.'),
        default=False,
        db_index=True,
    )
    is_available = models.BooleanField(
        _('available'), default=False, db_index=True,
        help_text=_('Is still available for purchase')
    )
    is_visible = models.BooleanField(
        _('visible'), default=True, db_index=True,
        help_text=_('Is visible in pricing page')
    )
    quotas = models.ManyToManyField(Quota, through=PackageQuota)
    created = models.DateTimeField(_('created'), auto_now_add=True)
    modified = models.DateTimeField(_('modified'), auto_now=True)
    objects = PackageQuerySet.as_manager()
    i18n = TranslationField(fields=('title', 'description',))

    class Meta:
        verbose_name = _('package')
        verbose_name_plural = _('packages')
        ordering = ['order']
        constraints = [
            models.UniqueConstraint(fields=['is_default'], condition=models.Q(is_default=True), name='unique_default'),
            models.UniqueConstraint(fields=['is_fallback'], condition=models.Q(is_fallback=True), name='unique_fallback'),
        ]

    def __str__(self):
        return self.title_i18n

    def get_absolute_url(self):
        return reverse('fees:packages')

    @classmethod
    def get_default_package(cls):
        try:
            return_value = cls.objects.get(is_default=True)
        except cls.DoesNotExist:
            return_value = None
        return return_value

    @classmethod
    def get_fallback_package(cls):
        try:
            return_value = cls.objects.get(is_fallback=True)
        except cls.DoesNotExist:
            return_value = None
        return return_value

    @classmethod
    def get_current_package(cls, purchaser):
        """
        Get current package for purchaser.
        App should be responsible for creating purchaser's plan (after sign up for example).
        If plan is expired or not present, return fallback package if available else None.
        """
        # We need to handle both default package (new purchaser -> TRIAL) and expired plan -> fallback

        is_anonymous = isinstance(purchaser, get_user_model()) and purchaser.is_anonymous

        # anonymous user
        if not purchaser or is_anonymous:
            # TODO: any other rules?
            return None

        plan_is_set = hasattr(purchaser, 'plan') and purchaser.plan is not None

        # currently, valid (not expired) package
        if plan_is_set and not purchaser.plan.is_expired():
            return purchaser.plan.package

        # fallback package
        fallback_package = Package.get_fallback_package()

        # validation of fallback package price (fallback package can't be free)
        if fallback_package is not None and not fallback_package.is_free():
            if plan_is_set:
                raise ValidationError(_('Plan has expired and fallback package is not free'))
            else:
                raise ValidationError(_('Plan not found and fallback package is not free'))

        # fallback package for expired or not present plan
        return fallback_package

    def get_quotas(self):
        quota_dic = {}
        for package_quota in PackageQuota.objects.filter(package=self).select_related('quota'):
            if Package.is_quota_available(package_quota):
                quota_dic[package_quota.quota.codename] = package_quota.value
        return quota_dic

    @staticmethod
    def is_quota_available(package_quota):
        value = package_quota['value'] if isinstance(package_quota, dict) else package_quota.value
        return value is None or value > 0

    def is_free(self):
        return not self.pricing_set.exists()
    is_free.boolean = True

    def monthly_plan(self):
        return self.pricing_set.filter(period=Pricing.PERIOD_MONTH).order_by('duration').first()

    is_free.boolean = True


class Pricing(models.Model):
    PERIOD_DAY = 'DAY'
    PERIOD_MONTH = 'MONTH'
    PERIOD_YEAR = 'YEAR'
    PERIODS = [
        (PERIOD_DAY, _('day')),
        (PERIOD_MONTH, _('month')),
        (PERIOD_YEAR, _('year')),
    ]
    PERIODS_PLURALIZE = [
        (PERIOD_DAY, (_('day'), _('days'))),
        (PERIOD_MONTH, (_('month'), _('months'))),
        (PERIOD_YEAR, (_('year'), _('years'))),
    ]

    package = models.ForeignKey(Package, on_delete=models.CASCADE)
    period = models.CharField(_('period'), choices=PERIODS, max_length=5)
    duration = models.PositiveSmallIntegerField(verbose_name=_('duration'), help_text=_('in period'),
                                           blank=True, null=True, default=None)
    price = models.DecimalField(_('price'), help_text=fees_settings.CURRENCY, max_digits=10, decimal_places=2, db_index=True, validators=[MinValueValidator(0.01)])
    # is_default = models.BooleanField(
    #     help_text=_('Default pricing for package'),
    #     default=False,
    #     db_index=True,
    # )
    created = models.DateTimeField(_('created'), auto_now_add=True)
    modified = models.DateTimeField(_('modified'), auto_now=True)
    # objects = PricingQuerySet.as_manager()

    class Meta:
        verbose_name = _('pricing')
        verbose_name_plural = _('pricings')
        ordering = ['price']
        unique_together = (['package', 'period', 'duration'],)

    @classmethod
    def distinct_period_durations(cls):
        period_durations = cls.objects.order_by('period', 'duration').values('period', 'duration').distinct()

        for period_duration in period_durations:
            period_duration['display'] = Pricing.period_duration_display(
                period=period_duration['period'],
                duration=period_duration['duration']
            )

        return period_durations

    @staticmethod
    def period_duration_display(period, duration):
        period_localize = dict(Pricing.PERIODS_PLURALIZE).get(period)
        period_display = period_localize[0] if duration == 1 else period_localize[1]  # TODO: i18n
        return f'{duration} {period_display}'

    def __str__(self):
        return f'{self.package} ({self.get_duration_display()})'

    def get_absolute_url(self):
        return reverse('fees:pricing')

    def get_duration_display(self):
        return Pricing.period_duration_display(self.period, self.duration)

    def get_price_display(self):
        return f'{self.price} {fees_settings.CURRENCY}'

    # TODO: remove commerce
    def get_add_to_cart_url(self):  # TODO: move to ProductMixin, template tag?
        if 'commerce' in settings.INSTALLED_APPS:
            content_type = ContentType.objects.get_for_model(self)
            return reverse('commerce:add_to_cart', args=(content_type.id, self.id))  # TODO: human readable URL
        return '#'

    # TODO: remove commerce, add product availability handler to commerce
    # @property
    # def availability(self):
    #     from commerce.models import AbstractProduct
    #     return AbstractProduct.AVAILABILITY_DIGITAL_GOODS

    @property
    def price_per_month(self):
        price = self.price

        if self.period == self.PERIOD_DAY:
            price = self.price * 30 / self.duration  # approximately

        if self.period == self.PERIOD_MONTH:
            price = self.price / self.duration

        if self.period == self.PERIOD_YEAR:
            price = self.price / 12 / self.duration

        return round(price, 2)

    # TODO: remove commerce
    def get_price_per_month_display(self):
        return f'{self.price_per_month} {fees_settings.CURRENCY}'

    @property
    def timedelta(self):
        if self.period == self.PERIOD_DAY:
            return timedelta(days=self.duration)

        if self.period == self.PERIOD_MONTH:
            return relativedelta(months=self.duration)

        if self.period == self.PERIOD_YEAR:
            return relativedelta(years=self.duration)


class Plan(models.Model):
    """
    Currently selected plan for purchaser.
    """
    purchaser = models.ForeignKey(
        get_purchaser_model(), verbose_name=_('purchaser'),
        on_delete=models.CASCADE,
    )
    package = models.ForeignKey(Package, verbose_name=_('package'), on_delete=models.CASCADE)
    pricing = models.ForeignKey(Pricing, help_text=_('pricing'), default=None,
                                null=True, blank=True, on_delete=models.CASCADE)
    activation = models.DateField(_('activation'), auto_now_add=True)
    expiration = models.DateField(_('expires'), help_text=_('keep blank to calculate automatically by pricing period'),
        default=None, blank=True, null=True, db_index=True)
    # is_active = models.BooleanField(_('active'), default=True, db_index=True)
    # is_recurring = models.BooleanField(_('active'), default=True, db_index=True)  # TODO: can be turned on/turned off
    modified = models.DateTimeField(_('modified'), auto_now=True)
    objects = PlanQuerySet.as_manager()

    class Meta:
        verbose_name = _("Plan")
        verbose_name_plural = _("Plans")
        ordering = ['-activation']
        get_latest_by = 'expiration'

    def __str__(self):
        if self.pricing:
            return "%s [%s] (%s)" % (self.purchaser, self.package, self.pricing)
        return "%s [%s]" % (self.purchaser, self.package)

    def save(self, **kwargs):
        # update expiration date by pricing period
        if self.expiration in EMPTY_VALUES and self.pricing:
            from_date = self.activation or now().date()  # TODO: now().date() vs date.today()
            self.expiration = from_date + self.pricing.timedelta
        return super().save(**kwargs)

    # def is_active(self):
    #     return self.active

    def is_expired(self):
        if self.expiration is None:
            return False
        else:
            return self.expiration < date.today()

    @property
    def days_left(self):
        if self.expiration is None:
            return None
        else:
            return (self.expiration - date.today()).days

    def expires_soon(self):
        days_left = self.days_left
        return days_left is not None and days_left <= 7  # TODO: add to settings

    # def clean_activation(self):
    #     errors = plan_validation(self.purchaser)
    #     if not errors['required_to_activate']:
    #         plan_validation(self.purchaser, on_activation=True)
    #         self.activate()
    #     else:
    #         self.deactivate()
    #     return errors
    #
    # def activate(self):
    #     if not self.active:
    #         self.active = True
    #         self.save()
    #         account_activated.send(sender=self, purchaser=self.purchaser)
    #
    # def deactivate(self):
    #     if self.active:
    #         self.active = False
    #         self.save()
    #         account_deactivated.send(sender=self, purchaser=self.purchaser)

    # def initialize(self):
    #     """
    #     Set up purchaser plan for first use
    #     """
    #     if not self.is_active():
    #         # Plans without pricings don't need to expire
    #         if self.expiration is None and self.plan.pricing_set.count():
    #             self.expiration = now() + timedelta(
    #                 days=getattr(settings, 'PLANS_DEFAULT_GRACE_PERIOD', 30))
    #         self.activate()  # this will call self.save()

    def get_extended_from(self, package):
        if package.is_free():
            return None
        if not self.is_expired() and self.expiration is not None and self.package == package:
            return self.expiration
        return date.today()

    def get_extended_until(self, package, pricing):
        if package.is_free():
            return None
        # if not self.package.is_free() and self.expiration is None:
        #     return None  # TODO: why? which use case?
        # if pricing is None:  # TODO: why? which use case?
        #     return self.expiration
        # return self.get_extended_from(package) + pricing.timedelta
        from_date = self.get_extended_from(package)
        return from_date + pricing.timedelta

    def plan_autorenew_at(self):
        """
        Helper function which calculates when the plan autorenewal will occur
        """
        if self.expiration:
            plans_autorenew_before_days = getattr(settings, 'FEES_AUTORENEW_BEFORE_DAYS', 0)
            plans_autorenew_before_hours = getattr(settings, 'FEES_AUTORENEW_BEFORE_HOURS', 0)
            return self.expiration - timedelta(days=plans_autorenew_before_days, hours=plans_autorenew_before_hours)

    # def set_plan_renewal(self, order, has_automatic_renewal=True, **kwargs):
    #     """
    #     Creates or updates plan renewal information for this plan with given order
    #     """
    #     if hasattr(self, 'recurring'):
    #         # Delete the plan to populate with default values
    #         # We don't want to mix the old and new values
    #         self.recurring.delete()
    #     recurring = RecurringPlan.objects.create(
    #         plan=self,
    #         pricing=order.pricing,
    #         amount=order.amount,
    #         tax=order.tax,
    #         currency=order.currency,
    #         has_automatic_renewal=has_automatic_renewal,
    #         **kwargs,
    #     )
    #     return recurring

    def extend(self, package, pricing):
        """
        Manages extending plan after package/pricing order
        :param package:
        :param pricing: if pricing is None then plan will be only upgraded
        :return:
        """

        if pricing and pricing.package != package:
            raise ValueError(f'Extending by package {package} by invalid pricing {pricing}!')

        status = False  # flag; if extending account was successful?
        new_expiration = self.get_extended_until(package, pricing)

        if pricing is None:
            # Process a plan change request (downgrade or upgrade)
            # No account activation or extending at this point
            self.package = package

            # if self.expiration is not None and not plan.pricing_set.count():
            #     # Assume no expiry date for plans without pricing.
            #     self.expiration = None

            self.expiration = new_expiration
            self.save()

            purchaser_change_package.send(sender=self, purchaser=self.purchaser)
            # if getattr(settings, 'FEES_SEND_EMAILS_PACKAGE_CHANGED', True):
            #     mail_context = {'purchaser': self.purchaser, 'plan': self, 'plan': plan}
            #     send_template_email([self.purchaser.email], 'mail/change_plan_title.txt', 'mail/change_plan_body.txt',
            #                         mail_context, get_user_language(self.purchaser))
            logger.info(
                "Account '%s' [id=%d] package changed to '%s' [id=%d]" % (self.purchaser, self.purchaser.pk, package, package.pk))
            status = True

        else:
            # Processing standard account extending procedure
            if self.package == package:
                status = True
                if self.pricing != pricing:
                    self.pricing = pricing
            else:
                # This should not ever happen (as this case should be managed by plan change request)
                # but just in case we consider a case when user has a different plan
                if not self.package.is_free() and self.expiration is None:
                    status = True
                elif not self.package.is_free() and self.expiration > date.today():
                    status = False
                    logger.warning("Purchaser '%s' [id=%d] package NOT changed to '%s' [id=%d]" % (
                        self.purchaser, self.purchaser.pk, package, package.pk))
                else:
                    status = True
                    purchaser_change_package.send(sender=self, purchaser=self.purchaser)
                    self.package = package
                    self.pricing = pricing

            if status:
                self.expiration = new_expiration
                self.save()
                logger.info("Purchaser '%s' [id=%d] has been extended by %d days using package '%s' [id=%d]" % (
                    self.purchaser, self.purchaser.pk, pricing.timedelta.days, package, package.pk))
                # if getattr(settings, 'PLANS_SEND_EMAILS_PLAN_EXTENDED', True):
                # mail_context = {'purchaser': self.purchaser,
                #                 'plan': self,
                #                 'package': package,
                #                 'pricing': pricing}
                # TODO: get purchaser email
                # send_template_email([self.purchaser.email], 'mail/extend_account_title.txt',
                #                     'mail/extend_account_body.txt',
                #                     mail_context, get_user_language(self.user))

        # if status:
        #     self.clean_activation()

        return status

    # def expire(self):
    #     """manages expiration"""
    #
    #     self.deactivate()
    #
    #     logger.info(
    #         "Purchaser '%s' [id=%d] has expired" % (self.purchaser, self.purchaser.pk))
    #
    #     mail_context = {'purchaser': self.purchaser, 'plan': self}
    #     send_template_email([self.purchaser.email], 'mail/expired_account_title.txt', 'mail/expired_account_body.txt',
    #                         mail_context, get_user_language(self.purchaser))
    #
    #     account_expired.send(sender=self, user=self.purchaser)

    def send_reminder(self):
        if self.is_expired():
            return

        # TODO: get language of purchaser
        with override_language(self.purchaser.preferred_language):
            EmailManager.send_mail(self.purchaser, 'fees/mails/subscription_reminder', _('Your subscription is going to expire soon'), data={'plan': self}, request=None)

    # def remind_expire_soon(self):
    #     """reminds about soon account expiration"""
    #
    #     mail_context = {
    #         'purchaser': self.purchaser,
    #         'plan': self,
    #         'days': self.days_left
    #     }
    #     send_template_email([self.user.email], 'mail/remind_expire_title.txt', 'mail/remind_expire_body.txt',
    #                         mail_context, get_user_language(self.user))

    @classmethod
    def create_for_purchaser(cls, purchaser, pricing=None):
        if pricing:
            package = pricing.package
            expiration = now() + pricing.timedelta
        else:
            package = Package.get_default_package()

            # check if default package is available
            if not package:
                return
    
            expiration = None if package.is_free() else now() + timedelta(days=package.trial_duration)

        return Plan.objects.create(
            purchaser=purchaser,
            package=package,
            pricing=pricing,
            # active=False,
            expiration=expiration,
        )

    @classmethod
    def create_for_purchasers_without_plan(cls):
        purchasers = get_purchaser_model().objects.filter(plan=None).only('pk')

        for purchaser in purchasers:
            Plan.create_for_purchaser(purchaser)

        return purchasers


# class RecurringPlan(models.Model):
#     """
#     OneToOne model associated with Plan that stores information about the plan recurrence.
#     More about recurring payments in docs.
#     """
#     plan = models.OneToOneField('Plan', on_delete=models.CASCADE, related_name='recurring')
#     token = models.CharField(
#         _('recurring token'),
#         help_text=_('Token, that will be used for payment renewal. Depends on used payment provider'),
#         max_length=255,
#         default=None,
#         null=True,
#         blank=True,
#     )
#     payment_provider = models.CharField(
#         _('payment provider'),
#         help_text=_('Provider, that will be used for payment renewal'),
#         max_length=255,
#         default=None,
#         null=True,
#         blank=True,
#     )
#     pricing = models.ForeignKey('Pricing', help_text=_('Recurring pricing'), default=None,
#                                 null=True, blank=True, on_delete=models.CASCADE)
#     amount = models.DecimalField(
#         _('amount'), max_digits=7, decimal_places=2, db_index=True, null=True, blank=True)
#     tax = models.DecimalField(_('tax'), max_digits=4, decimal_places=2, db_index=True, null=True,
#                               blank=True)  # Tax=None is when tax is not applicable
#     currency = models.CharField(_('currency'), max_length=3)
#     has_automatic_renewal = models.BooleanField(
#         _('has automatic plan renewal'),
#         help_text=_(
#             'Automatic renewal is enabled for associated plan. '
#             'If False, the plan renewal can be still initiated by user.',
#         ),
#         default=False,
#     )
    # card_expire_year = models.IntegerField(null=True, blank=True)
    # card_expire_month = models.IntegerField(null=True, blank=True)
    # card_masked_number = models.CharField(null=True, blank=True, max_length=255)
    #
    # def create_renew_order(self):
    #     """
    #     Create order for plan renewal
    #     """
    #     plan = self.user_plan
    #     return Order.objects.create(
    #         user=plan.user,
    #         plan=plan.plan,
    #         pricing=plan.recurring.pricing,
    #         amount=plan.recurring.amount,
    #         tax=plan.recurring.tax,
    #         currency=plan.recurring.currency,
    #     )

# class PlanPricingManager(models.Manager):
#     def get_query_set(self):
#         return super(PlanPricingManager, self).get_query_set().select_related('plan', 'pricing')


# class PlanPricing(models.Model):
#     plan = models.ForeignKey('Plan', on_delete=models.CASCADE)
#     pricing = models.ForeignKey('Pricing', on_delete=models.CASCADE)
#     price = models.DecimalField(max_digits=7, decimal_places=2, db_index=True)
#     order = models.IntegerField(default=0, null=False, blank=False)
#     has_automatic_renewal = models.BooleanField(
#         _('has automatic renewal'),
#         help_text=_('Use automatic renewal if possible?'),
#         default=False,
#     )
#
#     objects = PlanPricingManager()
#
#     class Meta:
#         ordering = ('order', 'pricing__period', )
#         verbose_name = _("Plan pricing")
#         verbose_name_plural = _("Plans pricings")
#
#     def __str__(self):
#         return "%s %s" % (self.plan.name, self.pricing)

from .signals import *



# Generated by Django 3.1.14 on 2022-07-06 14:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('fees', '0005_auto_20220706_1625'),
    ]

    operations = [
        migrations.AddConstraint(
            model_name='package',
            constraint=models.UniqueConstraint(condition=models.Q(is_fallback=True), fields=('is_fallback',), name='unique_fallback'),
        ),
    ]



# Generated by Django 3.1.14 on 2022-04-04 19:16

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import modeltrans.fields

from fees import settings as fees_settings


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(fees_settings.PURCHASER_MODEL),
    ]

    for dependency in fees_settings.MIGRATION_DEPENDENCIES:
        dependencies.append(dependency)

    operations = [
        migrations.CreateModel(
            name='Package',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=50, unique=True)),
                ('order', models.PositiveSmallIntegerField(default=1, help_text='to set order in pricing', unique=True, verbose_name='ordering')),
                ('trial_duration', models.PositiveSmallIntegerField(default=0, help_text='in days', verbose_name='trial duration')),
                ('is_default', models.BooleanField(db_index=True, default=False, help_text='Default package for purchaser')),
                ('is_available', models.BooleanField(db_index=True, default=False, help_text='Is still available for purchase', verbose_name='available')),
                ('is_visible', models.BooleanField(db_index=True, default=True, help_text='Is visible in current offer', verbose_name='visible')),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', models.DateTimeField(auto_now=True, verbose_name='modified')),
            ],
            options={
                'verbose_name': 'package',
                'verbose_name_plural': 'packages',
                'ordering': ['order'],
            },
        ),
        migrations.CreateModel(
            name='Quota',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('codename', models.CharField(db_index=True, max_length=50, unique=True, verbose_name='codename')),
                ('name', models.CharField(max_length=100, verbose_name='name')),
                ('unit', models.CharField(blank=True, max_length=100, verbose_name='unit')),
                ('description', models.TextField(blank=True, verbose_name='description')),
                ('is_boolean', models.BooleanField(default=False, verbose_name='is boolean')),
                ('i18n', modeltrans.fields.TranslationField(fields=('name', 'description'), required_languages=(), virtual_fields=True)),
            ],
            options={
                'verbose_name': 'Quota',
                'verbose_name_plural': 'Quotas',
                'ordering': ('name',),
            },
        ),
        migrations.CreateModel(
            name='Pricing',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('period', models.CharField(choices=[('DAY', 'day'), ('MONTH', 'month'), ('YEAR', 'year')], max_length=5, verbose_name='period')),
                ('duration', models.PositiveSmallIntegerField(blank=True, default=None, help_text='in period', null=True, verbose_name='duration')),
                ('price', models.DecimalField(db_index=True, decimal_places=2, help_text='EUR', max_digits=10, validators=[django.core.validators.MinValueValidator(0.01)], verbose_name='price')),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', models.DateTimeField(auto_now=True, verbose_name='modified')),
                ('package', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='fees.package')),
            ],
            options={
                'verbose_name': 'pricing',
                'verbose_name_plural': 'pricing plans',
                'ordering': ['price'],
            },
        ),
        migrations.CreateModel(
            name='Plan',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('expiration', models.DateField(blank=True, db_index=True, default=None, null=True, verbose_name='expires')),
                ('modified', models.DateTimeField(auto_now=True, verbose_name='modified')),
                ('package', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='fees.package', verbose_name='package')),
                ('pricing', models.ForeignKey(blank=True, default=None, help_text='pricing', null=True, on_delete=django.db.models.deletion.CASCADE, to='fees.pricing')),
                ('purchaser', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='plan', related_query_name='plan', to=fees_settings.PURCHASER_MODEL, verbose_name='purchaser')),
            ],
            options={
                'verbose_name': 'Plan',
                'verbose_name_plural': 'Plans',
            },
        ),
        migrations.CreateModel(
            name='PackageQuota',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.IntegerField(blank=True, default=1, null=True)),
                ('package', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='fees.package')),
                ('quota', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='fees.quota')),
            ],
            options={
                'verbose_name': 'Package quota',
                'verbose_name_plural': 'Packages quotas',
            },
        ),
        migrations.AddField(
            model_name='package',
            name='quotas',
            field=models.ManyToManyField(through='fees.PackageQuota', to='fees.Quota'),
        ),
    ]



from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.db.models.signals import post_save, pre_save
from django.dispatch import Signal, receiver

# TODO: commerce
# from commerce.models import Cart, Order
# from commerce.signals import cart_updated
from fees.helpers import get_purchaser_model
from fees.models import Plan, Pricing
from pragmatic.signals import apm_custom_context, SignalsHelper

User = get_user_model()


order_started = Signal()
order_started.__doc__ = """
Sent after order was started (awaiting payment)
"""

order_completed = Signal()
order_completed.__doc__ = """
Sent after order was completed (payment accepted, account extended)
"""


user_language = Signal(providing_args=['user', 'language'])
user_language.__doc__ = """Sent to receive information about language for user account"""

account_automatic_renewal = Signal(providing_args=['user'])
account_automatic_renewal.__doc__ = """
Try to renew the account automatically.
Should renew the user's UserPlan by recurring payments. If this succeeds, the plan should be extended.
"""

account_expired = Signal(providing_args=['user'])
account_expired.__doc__ = """
Sent on account expiration.
This signal is send regardless ``account_deactivated``
it only means that account has expired due to plan expire date limit.
"""

account_deactivated = Signal(providing_args=['user'])
account_deactivated.__doc__ = """
Sent on account deactivation, account is not operational (it could be not expired, but does not meet quota limits).
"""

account_activated = Signal(providing_args=['user'])
account_activated.__doc__ = """
Sent on account activation, account is now fully operational.
"""
purchaser_change_package = Signal(providing_args=['user'])
purchaser_change_package.__doc__ = """
Sent on account when plan was changed after order completion
"""

activate_purchaser_plan = Signal(providing_args=['user'])
activate_purchaser_plan.__doc__ = """
This signal should be called when user has succesfully registered (e.g. he activated account via e-mail activation).
If you are using django-registration there is no need to call this signal.
"""


@receiver(post_save, sender=get_purchaser_model())
def set_default_purchaser_plan(sender, instance, created, **kwargs):
    """
    Creates default plan for the new user but also extending an account for default grace period.
    """

    if created:
        Plan.create_for_purchaser(instance)


# TODO: check
# Hook to django-registration to initialize plan automatically after user has confirm account

@receiver(activate_purchaser_plan)
def initialize_plan_generic(sender, purchaser, **kwargs):
    try:
        purchaser.plan.initialize()
    except Plan.DoesNotExist:
        return


# @receiver(cart_updated, sender=Cart)
# @apm_custom_context('signals')
# def new_item_in_cart(sender, item, **kwargs):
#     # if new plan added into cart, delete all other items
#     if isinstance(item.product, Pricing):
#         item.cart.item_set.exclude(id=item.id).delete()


# @receiver(pre_save, sender=Order)
# @apm_custom_context('signals')
# def order_status_changed(sender, instance, **kwargs):
#     if instance.pk and SignalsHelper.attribute_changed(instance, ['status']) and instance.status == Order.STATUS_PAYMENT_RECEIVED:
#         if instance.has_item_of_type(Pricing):
#             purchased_pricings = instance.items_of_type(Pricing)
#
#             if purchased_pricings.count() > 1:
#                 raise ValueError(f'Order {instance.number} contains multiple subscription pricing plans!')
#
#             purchased_pricing = purchased_pricings.first()  # order should have single pricing plan only actually
#
#             purchaser = instance.user
#             pricing = purchased_pricing.product
#             plan = pricing.plan
#
#             try:
#                 # extend (and upgrade if necessary) plan
#                 purchaser.plan.extend(package, pricing)
#             except ObjectDoesNotExist:
#                 # create new plan
#                 Plan.create_for_user(instance.user, pricing)
